"""
Tools from the Attosecond science group at the Max Planck Institute of Quantum Optics, a.k.a. [Attoworld](https://www.attoworld.de)

## Structure
The module has several submodules to keep it organized:

- *file*: functions for reading/writing the various data formats used across the labs

- *numeric*: numerical tools

- *personal*: a module where we can add our own functions that might not be of general interest, but are still good to have available to we can easily share work

- *plot*: functions for plotting with a consistent style

- *wave*: functions for processing waveforms

- *spectrum*: functions for processing spectra

## Guidelines
The goal of this module is to gather the python programming work that we do, which maybe others in the group or the community at large could benefit from, into a module that we can easily add to our projects. This is easier if we follow some guidelines for best practices:

 - Use [docstrings](https://google.github.io/styleguide/pyguide.html#38-comments-and-docstrings) so that people know how to use your functions.

 - Comment code enough that it's understandable. It's possible that you write in a "self documenting" way, which is fine, but if you're doing something fancy and non-obvious, put in a note

 - If there's a function or class that you think others could benefit from, absolutely feel free to add it to the main modules. If you think you are likely the only one who will use something, you can also add a submodule to the attoworld.personal namespace. This makes it easier to share files with others!

"""

from . import file
from . import numeric
from . import personal
from . import plot
from . import spectrum
from . import wave



import numpy as np

def fornberg_stencil(order: int, positions: np.ndarray, position_out: float = 0.0) -> np.ndarray:
    """
    Generate a finite difference stencil using the algorithm described by B. Fornberg
    in Mathematics of Computation 51, 699-706 (1988).

    Args:
        order (int): the order of the derivative
        positions (np.ndarray): the positions at which the functions will be evaluated in the stencil.
    Returns:
        np.ndarray: the finite difference stencil with weights corresponding to the positions in the positions input array

    Examples:

        >>> stencil = fornberg_stencil(1, [-1,0,1])
        >>> print(stencil)
        [-0.5 0. 0.5]
    """
    # Contributed by Nick Karpowicz
    number_of_positions = len(positions)
    delta = np.zeros((number_of_positions, number_of_positions, order + 1), dtype=float)
    delta[0, 0, 0] = 1.0
    c1 = 1.0

    for n in range(1, number_of_positions):
        c2 = 1.0
        for v in range(n):  # v from 0 to n-1
            c3 = positions[n] - positions[v]
            c2 *= c3
            if n <= order:
                delta[n-1, v, n] = 0.0
            for m in range(min(n, order) + 1):
                if m == 0:
                    last_element = 0.0
                else:
                    last_element = m * delta[n-1, v, m-1]
                delta[n, v, m] = ((positions[n] - position_out) * delta[n-1, v, m] - last_element) / c3
        for m in range(min(n, order) + 1):
            if m == 0:
                first_element = 0.0
            else:
                first_element = m * delta[n-1, n-1, m-1]
            delta[n, n, m] = (c1 / c2) * (first_element - (positions[n-1] - position_out) * delta[n-1, n-1, m])
        c1 = c2

    return delta[-1, :, -1].squeeze()

def uniform_derivative(data: np.ndarray, order: int = 1, neighbors: int = 1, boundary: str = 'internal') -> np.ndarray:
    """
    Use a Fornberg stencil to take a derivative of arbitrary order and accuracy, handling the edge
    by using modified stencils that only use internal points.

    Args:
        data (np.ndarray): the data whose derivative should be taken
        order (int): the order of the derivative
        neighbors (int): the number of nearest neighbors to consider.
        boundary (str): How to treat the boundary: 'internal' will use only internal points (default). 'periodic' will assume periodic boundary. 'zero' will assume the data is zero outsize the grid.

    Returns:
        np.ndarray: the derivative
    """

    positions = np.array(range(-neighbors,neighbors+1))
    stencil = fornberg_stencil(order, positions)
    derivative = np.convolve(data, np.flip(stencil), mode='same')

    match boundary:
        case 'zero':
            return derivative
        case 'periodic':
            boundary_array = np.concatenate((data[-2*neighbors::],data[0:(2*neighbors + 1)]))
            for _i in range(2*neighbors + 1):
                derivative[-neighbors + _i] = np.sum(boundary_array[_i:(_i + 2*neighbors + 1)] * stencil)
            return derivative
        case 'internal':
            # increase number of included neighbors to improve accuracy
            neighbors += 1
            positions = np.array(range(-neighbors,neighbors+1))
            def corrected_point_top(index: int):
                boundary_stencil = fornberg_stencil(order, positions + neighbors - index)
                return np.sum(boundary_stencil*data[0:len(positions)])

            def corrected_point_bottom(index: int):
                boundary_stencil = fornberg_stencil(order, positions-neighbors+index)
                return np.sum(boundary_stencil*data[(-len(positions))::])

            for _i in range(neighbors):
                derivative[_i] = corrected_point_top(_i)
                derivative[-1 -_i] = corrected_point_bottom(_i)

    return derivative

def interpolate(x_out:np.ndarray, x_in: np.ndarray, y_in:np.ndarray, neighbors: int = 2, extrapolate: bool = False) -> np.ndarray:
    """
    Use a Fornberg stencil containing a specified number of neighboring points to perform interpolation.

    Contributed by Nick Karpowicz

    Args:
        x_out (np.ndarray): array of output x values, the array onto which y_in will be interpolated
        x_in (np.ndarray): array of input x values
        y_in (np.ndarray): array of input y values
        neighbors (int): number of nearest neighbors to include in the interpolation
        extrapolate (bool): unless set to true, values outside of the range of x_in will be zero

    Returns:
        np.ndarray: the interpolated y_out
    """
    sort_order = np.argsort(x_in)
    x_in_sorted = x_in[sort_order]
    y_in_sorted = y_in[sort_order]
    y_out = np.zeros(x_out.shape)
    locations = np.searchsorted(x_in_sorted, x_out, side='left')

    def interpolate_front_edge(x):
        stencil = fornberg_stencil(
            order = 0,
            positions = x_in_sorted[0:(2*neighbors)],
            position_out=x)
        return np.sum(stencil * y_in_sorted[0:(2*neighbors)])

    def interpolate_rear_edge(x):
        stencil = fornberg_stencil(
            order = 0,
            positions = x_in_sorted[(-1 - 2*neighbors)::],
            position_out=x)
        return np.sum(stencil * y_in_sorted[(-1 - 2*neighbors)::])

    def interpolate_point(x, location):

        if (((location == 0) and x != x_in_sorted[0]) or (location >= len(x_in_sorted))):
            #points outside the range of x_in (extrapolation)
            if extrapolate:
                if location == 0:
                    return interpolate_front_edge(x)
                else:
                    return interpolate_rear_edge(x)
            else:
                return 0.0

        elif x == x_in_sorted[location]:
            #case if x is exactly in the x_in array
            #if multiple points match, return average
            return np.mean(y_in_sorted[location==x_in_sorted])

        elif location < neighbors:
            #use modified front-edge interp
            return interpolate_front_edge(x)

        elif location >= (len(x_in_sorted) - neighbors):
            #use modified rear-edge interp
            return interpolate_rear_edge(x)
        else:
            #normal interior point
            stencil = fornberg_stencil(
                order = 0,
                positions = x_in_sorted[(location-neighbors):(location + neighbors)],
                position_out=x)
            return np.sum(stencil * y_in_sorted[(location-neighbors):(location + neighbors)])

    for _i in range(len(x_out)):
        y_out[_i] = interpolate_point(x_out[_i], locations[_i])

    return y_out



"""
This module will contain functions related to the processing of spectra.
"""
from .spectrum import *
from . import UVSpectrumAnalysis
from . import VISSpectrumAnalysis
__all__ = [name for name in globals() if callable(globals()[name]) and not name.startswith("_")]



import numpy as np
from scipy import constants
from typing import Optional
from ..numeric import interpolate
def wavelength_to_frequency(wavelengths_nm: np.ndarray, spectrum: np.ndarray, frequencies: Optional[np.ndarray]=None):
    """Convert a wavelength spectrum in W/nm into a frequency spectrum in W/THz

    Args:
        wavelength_nm (np.ndarray): the wavelengths included in the data, in nanometers
        spectrum (np.ndarray): the spectrum corresponding to the input wavelength scale
        frequencies: (optional) frequency vector for the output. If not specified, a vector will be calculated such that resolution and range are preserved.
    Returns:
        f: frequencies (Hz)
        scaled_spectrum: the frequency-domain spectrum
    """
    # Contributed by Nick Karpowicz
    input_frequencies = 1e9 * constants.speed_of_light/wavelengths_nm

    if frequencies is None:
        frequency_step = np.min(np.abs(np.diff(input_frequencies)))
        min_frequency = np.min(input_frequencies)
        max_frequency = np.max(input_frequencies)
        frequency_count = int(np.ceil(max_frequency-min_frequency)/frequency_step)
        frequencies = min_frequency + frequency_step * np.array(range(0,frequency_count),dtype=float)

    #apply Jakobian scaling and use W/THz
    scaled_spectrum = 1e12 * constants.speed_of_light * spectrum / (input_frequencies**2)

    indicies = np.argsort(input_frequencies)
    scaled_spectrum = interpolate(frequencies, input_frequencies[indicies], scaled_spectrum[indicies])

    return frequencies, scaled_spectrum

def transform_limited_pulse_from_spectrometer(wavelengths_nm: np.ndarray, spectrum: np.ndarray, gate_level: Optional[float]=None):
    """Calculates the transform-limited pulse corresponding to a spectrum

    Args:
        wavelength_nm: the wavelengths included in the data, in nanometers
        spectrum: the spectrum corresponding to the input wavelength scale
        gate_level: (optional) level, relative to the maximum at which to apply a gate to the spectrum. For example, with gate_level=0.01, values less than 1% of the maximum signal will be set to zero

    Returns:
        t: time vector (s)
        pulse: the pulse intensity vs. time
    """
    f,spec = wavelength_to_frequency(wavelengths_nm, spectrum)
    df = f[1]-f[0]
    t = np.fft.fftshift(np.fft.fftfreq(f.shape[0],d=df))
    gated_spectrum = np.array(spec)
    if gate_level is not None:
        gated_spectrum[spec<(gate_level * np.max(spec))] = 0.0

    pulse = np.fft.fftshift(
        np.abs(
            np.fft.ifft(
                np.sqrt(gated_spectrum))))**2

    return t,pulse
